@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -213,7 +213,10 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
             statement -> state == STATE_S9 ||
             statement -> state == STATE_S10 ||
             statement -> state == STATE_S11 ||
-            statement -> state == STATE_S12 )
+            statement -> state == STATE_S12 ||
+            statement -> state == STATE_S13 ||
+            statement -> state == STATE_S14 ||
+            statement -> state == STATE_S15 )
     	{
        	 	dm_log_write( __FILE__, 
                 __LINE__, 
@@ -225,8 +228,8 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                 ERROR_HY010, NULL,
                 src_descriptor -> connection -> environment -> requested_version );
 
-			function_return( SQL_HANDLE_DESC, target_descriptor, SQL_SUCCESS );
-        	return function_return( SQL_HANDLE_DESC, src_descriptor, SQL_ERROR );
+			function_return_nodrv( SQL_HANDLE_DESC, target_descriptor, SQL_SUCCESS );
+        	return function_return_nodrv( SQL_HANDLE_DESC, src_descriptor, SQL_ERROR );
     	}
 	}
 
@@ -249,18 +252,18 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                 ERROR_HY010, NULL,
                 target_descriptor -> connection -> environment -> requested_version );
 
-			function_return( SQL_HANDLE_DESC, target_descriptor, SQL_SUCCESS );
-        	return function_return( SQL_HANDLE_DESC, target_descriptor, SQL_ERROR );
+        	return function_return_nodrv( IGNORE_THREAD, src_descriptor, SQL_ERROR );
     	}
 	}
 
     /*
-     * if both descriptors are from the same connection the we can just
+     * if both descriptors are from the same driver then we can just
      * pass it on
      */
 
-    if ( src_descriptor -> connection == 
-            target_descriptor -> connection && 
+    if ( (src_descriptor -> connection == target_descriptor -> connection ||
+          !strcmp(src_descriptor -> connection -> dl_name,
+                  target_descriptor -> connection -> dl_name) ) && 
             CHECK_SQLCOPYDESC( src_descriptor -> connection ))
     {
         SQLRETURN ret;
@@ -277,7 +280,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
 
         if ( log_info.log_flag )
         {
-            sprintf( src_descriptor -> msg, 
+            sprintf( target_descriptor -> msg, 
                     "\n\t\tExit:[%s]",
                         __get_return_status( ret, s1 ));
 
@@ -285,11 +288,10 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
                     __LINE__, 
                     LOG_INFO, 
                     LOG_INFO, 
-                    src_descriptor -> msg );
+                    target_descriptor -> msg );
         }
 
-        thread_release( SQL_HANDLE_DBC, src_descriptor -> connection );
-        return function_return( IGNORE_THREAD, target_descriptor, ret );
+        return function_return( IGNORE_THREAD, target_descriptor, ret, DEFER_R3 );
     }
     else
     {
@@ -327,7 +329,7 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
             else
                 thread_release( SQL_HANDLE_ENV, src_descriptor -> connection -> environment );
 
-            return function_return( IGNORE_THREAD, target_descriptor, SQL_ERROR );
+            return function_return_nodrv( IGNORE_THREAD, target_descriptor, SQL_ERROR );
         }
 
         /*
@@ -517,6 +519,6 @@ SQLRETURN SQLCopyDesc( SQLHDESC source_desc_handle,
         else
             thread_release( SQL_HANDLE_ENV, src_descriptor -> connection -> environment );
 
-        return function_return( IGNORE_THREAD, target_descriptor, ret );
+        return function_return( IGNORE_THREAD, target_descriptor, ret, DEFER_R3 );
     }
 }
