@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -248,12 +248,13 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                         LOG_INFO, 
                         LOG_INFO, 
                         "Error: HY024" );
-        
+
+                function_entry( connection );
                 __post_internal_error( &connection -> error,
                     ERROR_HY024, NULL,
                     connection -> environment -> requested_version );
         
-                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
             else 
             {
@@ -317,12 +318,13 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                             LOG_INFO, 
                             LOG_INFO, 
                             "Error: HY024" );
-            
+
+                    function_entry( connection );
                     __post_internal_error( &connection -> error,
                         ERROR_HY024, NULL,
                         connection -> environment -> requested_version );
             
-                    return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
                 else 
                 {
@@ -347,13 +349,14 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                         __LINE__, 
                         LOG_INFO, 
                         LOG_INFO, 
-                        "Error: HY009" );
-        
+                        "Error: HY024" );
+
+                function_entry( connection );
                 __post_internal_error( &connection -> error,
-                    ERROR_HY009, NULL,
+                    ERROR_HY024, NULL,
                     connection -> environment -> requested_version );
         
-                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
             else 
             {
@@ -416,7 +419,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                 ERROR_08003, NULL,
                 connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
     else if ( connection -> state == STATE_C3 )
@@ -431,7 +434,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
             ERROR_HY010, NULL,
             connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
     else if ( connection -> state == STATE_C4 ||
             connection -> state == STATE_C5 ||
@@ -449,7 +452,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                 ERROR_08002, NULL,
                 connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
         else if ( attribute == SQL_ATTR_PACKET_SIZE )
         {
@@ -463,7 +466,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                 ERROR_HY011, NULL,
                 connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
     }
 
@@ -484,7 +487,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                 ERROR_HY024, NULL,
                 connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
     /*
@@ -493,10 +496,34 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
 
     switch( attribute ) 
     {
+        /* ODBC 3.x statement attributes are not settable at the connection level */
+        case SQL_ATTR_APP_PARAM_DESC:
+        case SQL_ATTR_APP_ROW_DESC:
+        case SQL_ATTR_CURSOR_SCROLLABLE:
+        case SQL_ATTR_CURSOR_SENSITIVITY:
+        case SQL_ATTR_ENABLE_AUTO_IPD:
+        case SQL_ATTR_FETCH_BOOKMARK_PTR:
+        case SQL_ATTR_IMP_PARAM_DESC:
+        case SQL_ATTR_IMP_ROW_DESC:
+        case SQL_ATTR_PARAM_BIND_OFFSET_PTR:
+        case SQL_ATTR_PARAM_BIND_TYPE:
+        case SQL_ATTR_PARAM_OPERATION_PTR:
+        case SQL_ATTR_PARAM_STATUS_PTR:
+        case SQL_ATTR_PARAMS_PROCESSED_PTR:
+        case SQL_ATTR_PARAMSET_SIZE:
+        case SQL_ATTR_ROW_ARRAY_SIZE:
+        case SQL_ATTR_ROW_BIND_OFFSET_PTR:
+        case SQL_ATTR_ROW_OPERATION_PTR:
+        case SQL_ATTR_ROW_STATUS_PTR:
+        case SQL_ATTR_ROWS_FETCHED_PTR:
+            __post_internal_error( &connection -> error,
+                    ERROR_HY092, NULL,
+                    connection -> environment -> requested_version );
+
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+
       	case SQL_ATTR_CONCURRENCY:
 	  	case SQL_BIND_TYPE:
-      	case SQL_ATTR_CURSOR_SCROLLABLE:
-      	case SQL_ATTR_CURSOR_SENSITIVITY:
       	case SQL_ATTR_CURSOR_TYPE:
       	case SQL_ATTR_MAX_LENGTH:
       	case SQL_MAX_ROWS:
@@ -507,11 +534,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
       	case SQL_ATTR_RETRIEVE_DATA:
       	case SQL_ATTR_SIMULATE_CURSOR:
       	case SQL_ATTR_USE_BOOKMARKS:
-            if( __check_stmt_from_dbc( connection, STATE_S8 ) ||
-                __check_stmt_from_dbc( connection, STATE_S9 ) ||
-                __check_stmt_from_dbc( connection, STATE_S10 ) ||
-                __check_stmt_from_dbc( connection, STATE_S11 ) ||
-                __check_stmt_from_dbc( connection, STATE_S12 )) {
+            if( __check_stmt_from_dbc_v( connection, 5, STATE_S8, STATE_S9, STATE_S10, STATE_S11, STATE_S12 )) {
 
                 dm_log_write( __FILE__, 
                         __LINE__, 
@@ -523,36 +546,13 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                         ERROR_24000, NULL,
                         connection -> environment -> requested_version );
 
-                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
             break;
 
         default:
-            if ( attribute == SQL_ATTR_CURRENT_CATALOG ) 
-            {
-                if( __check_stmt_from_dbc( connection, STATE_S5 ) ||
-                    __check_stmt_from_dbc( connection, STATE_S6 ) ||
-                    __check_stmt_from_dbc( connection, STATE_S7 )) {
 
-                    dm_log_write( __FILE__, 
-                            __LINE__, 
-                            LOG_INFO, 
-                            LOG_INFO, 
-                            "Error: 24000" );
-
-                    __post_internal_error( &connection -> error,
-                            ERROR_24000, NULL,
-                            connection -> environment -> requested_version );
-
-                    return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
-                }
-            }
-
-            if( __check_stmt_from_dbc( connection, STATE_S8 ) ||
-                __check_stmt_from_dbc( connection, STATE_S9 ) ||
-                __check_stmt_from_dbc( connection, STATE_S10 ) ||
-                __check_stmt_from_dbc( connection, STATE_S11 ) ||
-                __check_stmt_from_dbc( connection, STATE_S12 )) {
+            if( __check_stmt_from_dbc_v( connection, 5, STATE_S8, STATE_S9, STATE_S10, STATE_S11, STATE_S12 )) {
 
                 dm_log_write( __FILE__, 
                         __LINE__, 
@@ -564,7 +564,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                         ERROR_HY010, NULL,
                         connection -> environment -> requested_version );
 
-                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
             }
             break;
     }
@@ -576,7 +576,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
     value = __attr_override( connection, SQL_HANDLE_DBC, attribute, value, &string_length );
 
     /*
-     * we need to save this even if connected so we can use it for the next connetc
+     * we need to save this even if connected so we can use it for the next connect
      */
     if ( attribute == SQL_ATTR_LOGIN_TIMEOUT )
     {
@@ -661,12 +661,22 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
             }
             else if ( string_length == SQL_NTS )
             {
-                sa -> str_attr = strdup( value );
-                sa -> str_len = string_length;
+                if (!value)
+                {
+                    __post_internal_error( &connection -> error,
+                        ERROR_HY024, "Invalid argument value",
+                        connection -> environment -> requested_version );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                }
+                else
+                {
+                    sa -> str_attr = strdup( value );
+                    sa -> str_len = string_length;
+                }
             }
             else
             {
-                sa -> int_attr = ( SQLLEN ) value;
+                sa -> intptr_attr = (intptr_t) value;
                 sa -> str_len = string_length;
             }
             sa -> next = connection -> save_attr;
@@ -683,7 +693,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                 LOG_INFO, 
                 connection -> msg );
 
-        return function_return( SQL_HANDLE_DBC, connection, SQL_SUCCESS );
+        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_SUCCESS );
     }
     else
     {
@@ -710,7 +720,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                                 ERROR_HY092, NULL,
                                 connection -> environment -> requested_version );
 
-                        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                     }
 
                     ret = SQLSETCONNECTOPTION( connection,
@@ -730,7 +740,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
@@ -767,7 +777,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                                 ERROR_HY092, NULL,
                                 connection -> environment -> requested_version );
 
-                        return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                        return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                     }
 
                     switch( attribute )
@@ -775,7 +785,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                       case SQL_ATTR_CURRENT_CATALOG:
                       case SQL_ATTR_TRACEFILE:
                       case SQL_ATTR_TRANSLATE_LIB:
-                        s1 = ansi_to_unicode_alloc( value, SQL_NTS, connection );
+                        s1 = ansi_to_unicode_alloc( value, SQL_NTS, connection, NULL );
 
                         ret = SQLSETCONNECTOPTIONW( connection,
                                 connection -> driver_dbc,
@@ -807,7 +817,7 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                             ERROR_IM001, NULL,
                             connection -> environment -> requested_version );
 
-                    return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                    return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
                 }
             }
             else
@@ -819,13 +829,13 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
                   case SQL_ATTR_CURRENT_CATALOG:
                   case SQL_ATTR_TRACEFILE:
                   case SQL_ATTR_TRANSLATE_LIB:
-                    s1 = ansi_to_unicode_alloc( value, string_length, connection );
+                    s1 = ansi_to_unicode_alloc( value, string_length, connection, NULL );
 
                     ret = SQLSETCONNECTATTRW( connection,
                         connection -> driver_dbc,
                         attribute,
                         s1,
-                        string_length );
+                        string_length == SQL_NTS ? SQL_NTS : string_length * sizeof( SQLWCHAR ));
 
                     if ( s1 )
                         free( s1 );
@@ -865,5 +875,5 @@ SQLRETURN SQLSetConnectAttr( SQLHDBC connection_handle,
         connection -> bookmarks_on = (SQLLEN) value;
     }
 
-    return function_return( SQL_HANDLE_DBC, connection, ret );
+    return function_return( SQL_HANDLE_DBC, connection, ret, DEFER_R3 );
 }
