@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -205,13 +205,14 @@ SQLRETURN SQLSetScrollOptions(
                 ERROR_S1010, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
-    if ( crow_keyset != SQL_SCROLL_FORWARD_ONLY &&
+    if (( crow_keyset != SQL_SCROLL_FORWARD_ONLY &&
             crow_keyset != SQL_SCROLL_STATIC &&
             crow_keyset != SQL_SCROLL_KEYSET_DRIVEN &&
-            crow_keyset != SQL_SCROLL_DYNAMIC )
+            crow_keyset != SQL_SCROLL_DYNAMIC ) ||
+            !crow_rowset)
     {
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -223,7 +224,7 @@ SQLRETURN SQLSetScrollOptions(
                 ERROR_S1107, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     if ( f_concurrency != SQL_CONCUR_READ_ONLY &&
@@ -241,7 +242,7 @@ SQLRETURN SQLSetScrollOptions(
                 ERROR_S1108, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     if ( CHECK_SQLSETSCROLLOPTIONS( statement -> connection ))
@@ -252,7 +253,7 @@ SQLRETURN SQLSetScrollOptions(
                 crow_keyset,
                 crow_rowset );
     }
-    else if ( statement -> connection -> driver_act_ver == SQL_OV_ODBC3 &&
+    else if ( statement -> connection -> driver_act_ver >= SQL_OV_ODBC3 &&
             (CHECK_SQLGETINFO( statement -> connection ) || 
 			CHECK_SQLGETINFOW( statement -> connection )) &&
             (CHECK_SQLSETSTMTATTR( statement -> connection ) ||
@@ -295,7 +296,7 @@ SQLRETURN SQLSetScrollOptions(
                         ERROR_S1107, NULL,
                         statement -> connection -> environment -> requested_version );
 
-                return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+                return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
             }
             break;
         }
@@ -314,7 +315,7 @@ SQLRETURN SQLSetScrollOptions(
                 LOG_INFO, 
                 "Error: SQLGetInfo fails" );
             
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR, DEFER_R3 );
         }
 
         if ( f_concurrency == SQL_CONCUR_READ_ONLY &&
@@ -330,7 +331,7 @@ SQLRETURN SQLSetScrollOptions(
                     ERROR_S1C00, NULL,
                     statement -> connection -> environment -> requested_version );
             
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR, DEFER_R3 );
         }
         else if ( f_concurrency == SQL_CONCUR_LOCK &&
                 !( ivp & SQL_CA2_LOCK_CONCURRENCY ))
@@ -345,7 +346,7 @@ SQLRETURN SQLSetScrollOptions(
                     ERROR_S1C00, NULL,
                     statement -> connection -> environment -> requested_version );
             
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR, DEFER_R3 );
         }
         else if ( f_concurrency == SQL_CONCUR_ROWVER &&
                 !( ivp & SQL_CA2_OPT_ROWVER_CONCURRENCY ))
@@ -360,7 +361,7 @@ SQLRETURN SQLSetScrollOptions(
                     ERROR_S1C00, NULL,
                     statement -> connection -> environment -> requested_version );
             
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR, DEFER_R3 );
         }
         if ( f_concurrency == SQL_CONCUR_VALUES &&
                 !( ivp & SQL_CA2_OPT_VALUES_CONCURRENCY ))
@@ -375,7 +376,7 @@ SQLRETURN SQLSetScrollOptions(
                     ERROR_S1C00, NULL,
                     statement -> connection -> environment -> requested_version );
             
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR, DEFER_R3 );
         }
         if ( f_concurrency != SQL_CONCUR_READ_ONLY &&
             f_concurrency != SQL_CONCUR_LOCK &&
@@ -392,7 +393,7 @@ SQLRETURN SQLSetScrollOptions(
                     ERROR_S1108, NULL,
                     statement -> connection -> environment -> requested_version );
             
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR, DEFER_R3 );
         }
 
         if(CHECK_SQLSETSTMTATTR( statement -> connection ))
@@ -420,7 +421,7 @@ SQLRETURN SQLSetScrollOptions(
                 LOG_INFO, 
                 "Error: SQLSetStmtAttr fails" );
             
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR, DEFER_R3 );
         }
 
 
@@ -459,7 +460,7 @@ SQLRETURN SQLSetScrollOptions(
                         ERROR_S1107, NULL,
                         statement -> connection -> environment -> requested_version );
 
-                return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+                return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR, DEFER_R3 );
             }
             break;
         }
@@ -489,7 +490,7 @@ SQLRETURN SQLSetScrollOptions(
                 LOG_INFO, 
                 "Error: SQLSetStmtAttr fails" );
             
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR, DEFER_R3 );
         }
 
         if ( crow_keyset > 0 )
@@ -519,7 +520,7 @@ SQLRETURN SQLSetScrollOptions(
                     LOG_INFO, 
                     "Error: SQLSetStmtAttr fails" );
                 
-                return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+                return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR, DEFER_R3 );
             }
         }
         if(CHECK_SQLSETSTMTATTR( statement -> connection ))
@@ -551,7 +552,7 @@ SQLRETURN SQLSetScrollOptions(
                 ERROR_IM001, NULL,
                 statement -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
     }
 
     if ( log_info.log_flag )
@@ -567,5 +568,5 @@ SQLRETURN SQLSetScrollOptions(
                 statement -> msg );
     }
 
-    return function_return( SQL_HANDLE_STMT, statement, ret );
+    return function_return( SQL_HANDLE_STMT, statement, ret, DEFER_R3 );
 }
