@@ -1,7 +1,7 @@
 /*********************************************************************
  *
  * Written by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * copyright (c) 1999 Nick Gorham
  *
@@ -657,6 +657,7 @@ static attr_options env_options[] =
 		{
 			{ "SQL_OV_ODBC2", SQL_OV_ODBC2 }, 
 			{ "SQL_OV_ODBC3", SQL_OV_ODBC3 }, 
+			{ "SQL_OV_ODBC3_80", SQL_OV_ODBC3_80 }, 
 			{ NULL }
 		}, "3.0", SQL_INTEGER
 	},
@@ -734,6 +735,21 @@ int found = 0;
         opt ++;
     }
 
+    /*
+     * Handle non standard attributes by [numeric_value]={char value} or [numeric_value]=\int_value
+     */
+    if ( !found ) 
+    {
+        if ( kw[ 0 ] == '[' ) {
+            as -> attribute = atoi( kw + 1 );
+            if ( as -> value[ 0 ] == '\\' ) {
+                as -> is_int_type = 1;
+                as -> int_value = atoi( as -> value + 1 );
+            }
+            found = 1;
+        }
+    }
+
     return found;
 }
 
@@ -988,7 +1004,7 @@ static void __set_attribute( void *handle, int type, struct attr_set *as )
             return;
         }
 
-        if ( connection -> driver_version == SQL_OV_ODBC3 )
+        if ( connection -> driver_version >= SQL_OV_ODBC3 )
         {
             if ( CHECK_SQLSETENVATTR( connection ))
             {
@@ -1022,7 +1038,7 @@ static void __set_attribute( void *handle, int type, struct attr_set *as )
     {
         DMHDBC connection = (DMHDBC) handle;
 
-        if ( connection -> driver_version == SQL_OV_ODBC3 )
+        if ( connection -> driver_version >= SQL_OV_ODBC3 )
         {
             if ( CHECK_SQLSETCONNECTATTR( connection ))
             {
@@ -1094,7 +1110,7 @@ static void __set_attribute( void *handle, int type, struct attr_set *as )
         DMHSTMT statement = (DMHSTMT) handle;
         DMHDBC connection = statement -> connection;
 
-        if ( connection -> driver_version == SQL_OV_ODBC3 )
+        if ( connection -> driver_version >= SQL_OV_ODBC3 )
         {
             if ( CHECK_SQLSETSTMTATTR( connection ))
             {
@@ -1152,6 +1168,33 @@ static void __set_attribute( void *handle, int type, struct attr_set *as )
                             as -> value );
                 }
             }
+
+            /*
+             * Fall back, the attribute may be ODBC 3 only
+             */
+
+            if ( ret == SQL_ERROR ) {
+
+                if ( CHECK_SQLSETSTMTATTR( connection ))
+                {
+                    if ( as -> is_int_type )
+                    {
+                        ret = SQLSETSTMTATTR( connection,
+                                statement -> driver_stmt,
+                                as -> attribute,
+                                as -> int_value,
+                                0 );
+                    }
+                    else
+                    {
+                        ret = SQLSETSTMTATTR( connection,
+                                statement -> driver_stmt,
+                                as -> attribute,
+                                as -> value,
+                                strlen( as -> value ));
+                    }
+                }
+            }
         }
         if ( log_info.log_flag )
         {
@@ -1259,7 +1302,11 @@ void *__attr_override( void *handle, int type, int attribute, void *value, SQLIN
 
         if ( as -> is_int_type )
         {
+#ifdef HAVE_PTRDIFF_T
+            return (void*)(ptrdiff_t) as -> int_value;
+#else
             return (void*)(long) as -> int_value;
+#endif
         }
         else
         {
@@ -1321,7 +1368,11 @@ void *__attr_override_wide( void *handle, int type, int attribute, void *value,
 
         if ( as -> is_int_type )
         {
+#ifdef HAVE_PTRDIFF_T
+            return (void*)(ptrdiff_t) as -> int_value;
+#else
             return (void*)(long) as -> int_value;
+#endif
         }
         else
         {
@@ -1332,11 +1383,11 @@ void *__attr_override_wide( void *handle, int type, int attribute, void *value,
 			switch( type ) 
 			{
       			case SQL_HANDLE_DBC:
-					ansi_to_unicode_copy( buffer, as->value, SQL_NTS, (DMHDBC) handle );
+					ansi_to_unicode_copy( buffer, as->value, SQL_NTS, (DMHDBC) handle, NULL );
 					break;
 
       			case SQL_HANDLE_STMT:
-					ansi_to_unicode_copy( buffer, as->value, SQL_NTS, ((DMHSTMT) handle ) -> connection );
+					ansi_to_unicode_copy( buffer, as->value, SQL_NTS, ((DMHSTMT) handle ) -> connection, NULL );
 					break;
 			}
             return buffer;
@@ -1354,9 +1405,17 @@ void *__attr_override_wide( void *handle, int type, int attribute, void *value,
 
 int dm_check_connection_attrs( DMHDBC connection, SQLINTEGER attribute, SQLPOINTER value )
 {
+#ifdef HAVE_PTRDIFF_T
+    ptrdiff_t ival;
+#else
     SQLINTEGER ival;
+#endif
 
+#ifdef HAVE_PTRDIFF_T
+    ival = (ptrdiff_t) value;
+#else
     ival = (SQLINTEGER) value;
+#endif
 
     switch( attribute ) 
     {
@@ -1410,6 +1469,14 @@ int dm_check_connection_attrs( DMHDBC connection, SQLINTEGER attribute, SQLPOINT
             }
             break;
 
+        case SQL_ATTR_TXN_ISOLATION:
+            if ( ival != SQL_TXN_READ_UNCOMMITTED && ival != SQL_TXN_READ_COMMITTED
+                && ival != SQL_TXN_REPEATABLE_READ && ival != SQL_TXN_SERIALIZABLE )
+            {
+                return SQL_ERROR;
+            }
+            break;
+
         /*
          * include statement attributes as well
          */
@@ -1491,9 +1558,17 @@ int dm_check_connection_attrs( DMHDBC connection, SQLINTEGER attribute, SQLPOINT
 
 int dm_check_statement_attrs( DMHSTMT statement, SQLINTEGER attribute, SQLPOINTER value )
 {
+#ifdef HAVE_PTRDIFF_T
+    ptrdiff_t ival;
+#else
     SQLUINTEGER ival;
+#endif
 
+#ifdef HAVE_PTRDIFF_T
+    ival = (ptrdiff_t) value;
+#else
     ival = (SQLUINTEGER) value;
+#endif
 
     switch( attribute ) 
     {
@@ -1572,6 +1647,9 @@ int dm_check_statement_attrs( DMHSTMT statement, SQLINTEGER attribute, SQLPOINTE
             }
             break;
 
+        case SQL_ROWSET_SIZE:
+            return ival > 0 ? SQL_SUCCESS : SQL_ERROR;
+
         default:
             return SQL_SUCCESS;
     }
